#!/bin/env python
"""Reference Implementation of v1 C3 Security Beacon Auth Server."""
from aiohttp import web
import aiohttp_cors
import asyncio
import os
import signal

import click

from ..db import init_db, init_db_pool
from .listeners import ListenersView, ListenerView
from .proximity import ProximityView
from .tags import TagsView, TagView
from .zones import ZonesView, ZoneView
from .groups import GroupsView, GroupView
from ..demo_fe.main import app as demo_fe_app


@click.command()
@click.option('--port', '-p', default=5000)
@click.option('--reset-db', is_flag=True, default=False)
@click.option('--demo-fe', is_flag=True, default=True)
def main(port, reset_db, demo_fe) -> None:
    loop = asyncio.get_event_loop()
    if signal is not None and os.name != 'nt':
        loop.add_signal_handler(signal.SIGINT, loop.stop)

    async def init_db_app(app):
        pool = await init_db_pool()
        await init_db(pool, reset_db)
        app['db_pool'] = pool

    app = web.Application()
    app.on_startup.append(init_db_app)
    app.router.add_view('/api/listeners', ListenersView)
    app.router.add_view('/api/listeners/{id:\w+}', ListenerView, name='listener')
    app.router.add_view('/api/zones', ZonesView)
    app.router.add_view('/api/zones/{id:\d+}', ZoneView, name='zone')
    app.router.add_view('/api/tags', TagsView)
    app.router.add_view('/api/tags/{id:\d+}', TagView, name='tag')
    app.router.add_view('/api/proximity', ProximityView)
    app.router.add_view('/api/groups', GroupsView)
    app.router.add_view('/api/groups/{id:\d+}', GroupView, name='group')

    cors = aiohttp_cors.setup(app, defaults={
        "*": aiohttp_cors.ResourceOptions(
            allow_credentials=True,
            expose_headers="*",
            allow_headers="*",
        )
    })

    for route in list(app.router.routes()):
        cors.add(route)

    if demo_fe:
        app.add_subapp('/demo', demo_fe_app)
    web.run_app(app, port=port)


if __name__ == '__main__':
    main()



import lzo
import struct

import attr
from automat import MethodicalMachine

HEADER_LEN = 5


@attr.s
class RawMessage(object):
    data = attr.ib()
    type = attr.ib()


class Frame(object):
    _data_len = None
    _machine = MethodicalMachine()

    def __init__(self):
        self._data_len = 0
        self._compressed_p = False
        self._header = bytearray()
        self._data = bytearray()
        self._message_type = -1
        self.msg_buf = []

    @_machine.input()
    def add_bytes(self, data):
        """ Client calls this method to add rx'd bytes to the frame. """

    @_machine.input()
    def _have_header(self):
        """The header is complete."""

    @_machine.input()
    def _have_data(self):
        """The data is complete."""

    @_machine.input()
    def _have_compressed_data(self):
        """The data is complete, but compressed."""

    @_machine.input()
    def bad_frame(self):
        """ The frame is invalid. """

    @staticmethod
    def _save(data, dest, dest_len):
        cur = len(dest)
        remaining = dest_len - cur
        if len(data) >= remaining:
            dest.extend(data[:remaining])
            return True, data[remaining:]
        dest.extend(data)
        return False, None

    @_machine.output()
    def _save_header(self, data):
        done, remaining = self._save(data, self._header, HEADER_LEN)
        if done:
            self._have_header()
            self.add_bytes(remaining)

    @_machine.output()
    def _process_header(self):
        (self._data_len, flags) = struct.unpack("!IB", self._header[:5])
        self._compressed_p = True if flags >> 7 else False
        self.message_type = flags & ~(1 << 7)

    @_machine.output()
    def _save_data(self, data):
        done, remaining = self._save(data, self._data, self._data_len)
        if done:
            if self._compressed_p:
                self._have_compressed_data()
            else:
                self._have_data()
            if remaining:
                self.add_bytes(remaining)

    @_machine.state(initial=True)
    def _waiting_for_header(self):
        """In this state we don't have the header"""

    @_machine.state()
    def _waiting_for_data(self):
        """In this state we don't have all the data"""

    @_machine.output()
    def _decompress_data(self):
        try:
            d = lzo.decompress(bytes(self._data), False, 10240)  # 10kiB max payload
            self._data = d
            self._have_data()
        except lzo.error as e:
            print("Bad compressed frame: {}".format(e))
            self.bad_frame()

    @_machine.output()
    def _process_message(self):
        self.msg_buf.append(RawMessage(bytes(self._data), self.message_type))

    @_machine.output()
    def _clear(self):
        self._data = bytearray()
        self._header = bytearray()

    def get_messages(self):
        r = self.msg_buf
        self.msg_buf = []
        return r

    _waiting_for_header.upon(add_bytes, enter=_waiting_for_header,
                             outputs=[_save_header])
    _waiting_for_header.upon(_have_header, enter=_waiting_for_data,
                             outputs=[_process_header])
    _waiting_for_data.upon(add_bytes, enter=_waiting_for_data,
                           outputs=[_save_data])
    _waiting_for_data.upon(_have_compressed_data, enter=_waiting_for_data,
                           outputs=[_decompress_data])
    _waiting_for_data.upon(_have_data, enter=_waiting_for_header,
                           outputs=[_process_message, _clear])
    _waiting_for_data.upon(bad_frame, enter=_waiting_for_header,
                           outputs=[_clear])


def frame_message(data, compress=False):
    from c3agg.listener import MessageType
    flags = (1 << 7) if compress else 0
    flags |= MessageType.REQUEST
    data = lzo.compress(data, 6, False) if compress else data
    header = struct.pack("!IB", len(data), flags)
    return b''.join([header, data])



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: c3loc/proto/secbeacon.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='c3loc/proto/secbeacon.proto',
  package='',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x1b\x63\x33loc/proto/secbeacon.proto\"K\n\x10\x45ncryptedPayload\x12\n\n\x02id\x18\x01 \x01(\x0c\x12\r\n\x05nonce\x18\x02 \x01(\x0c\x12\x0f\n\x07payload\x18\x03 \x01(\x0c\x12\x0b\n\x03tag\x18\x04 \x01(\x0c\"\x12\n\x10PlaintextPayload\"\x9f\x01\n\x0fSecBeaconReport\x12\r\n\x05\x63lock\x18\x01 \x01(\x04\x12&\n\tencrypted\x18\x02 \x01(\x0b\x32\x11.EncryptedPayloadH\x00\x12&\n\tplaintext\x18\x03 \x01(\x0b\x32\x11.PlaintextPayloadH\x00\x12\x13\n\x0b\x64istance_cm\x18\x04 \x01(\r\x12\x10\n\x08variance\x18\x05 \x01(\rB\x06\n\x04\x64\x61tab\x06proto3'
)




_ENCRYPTEDPAYLOAD = _descriptor.Descriptor(
  name='EncryptedPayload',
  full_name='EncryptedPayload',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='EncryptedPayload.id', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='nonce', full_name='EncryptedPayload.nonce', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='payload', full_name='EncryptedPayload.payload', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='tag', full_name='EncryptedPayload.tag', index=3,
      number=4, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=31,
  serialized_end=106,
)


_PLAINTEXTPAYLOAD = _descriptor.Descriptor(
  name='PlaintextPayload',
  full_name='PlaintextPayload',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=108,
  serialized_end=126,
)


_SECBEACONREPORT = _descriptor.Descriptor(
  name='SecBeaconReport',
  full_name='SecBeaconReport',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='clock', full_name='SecBeaconReport.clock', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='encrypted', full_name='SecBeaconReport.encrypted', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='plaintext', full_name='SecBeaconReport.plaintext', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='distance_cm', full_name='SecBeaconReport.distance_cm', index=3,
      number=4, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='variance', full_name='SecBeaconReport.variance', index=4,
      number=5, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='data', full_name='SecBeaconReport.data',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=129,
  serialized_end=288,
)

_SECBEACONREPORT.fields_by_name['encrypted'].message_type = _ENCRYPTEDPAYLOAD
_SECBEACONREPORT.fields_by_name['plaintext'].message_type = _PLAINTEXTPAYLOAD
_SECBEACONREPORT.oneofs_by_name['data'].fields.append(
  _SECBEACONREPORT.fields_by_name['encrypted'])
_SECBEACONREPORT.fields_by_name['encrypted'].containing_oneof = _SECBEACONREPORT.oneofs_by_name['data']
_SECBEACONREPORT.oneofs_by_name['data'].fields.append(
  _SECBEACONREPORT.fields_by_name['plaintext'])
_SECBEACONREPORT.fields_by_name['plaintext'].containing_oneof = _SECBEACONREPORT.oneofs_by_name['data']
DESCRIPTOR.message_types_by_name['EncryptedPayload'] = _ENCRYPTEDPAYLOAD
DESCRIPTOR.message_types_by_name['PlaintextPayload'] = _PLAINTEXTPAYLOAD
DESCRIPTOR.message_types_by_name['SecBeaconReport'] = _SECBEACONREPORT
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

EncryptedPayload = _reflection.GeneratedProtocolMessageType('EncryptedPayload', (_message.Message,), {
  'DESCRIPTOR' : _ENCRYPTEDPAYLOAD,
  '__module__' : 'c3loc.proto.secbeacon_pb2'
  # @@protoc_insertion_point(class_scope:EncryptedPayload)
  })
_sym_db.RegisterMessage(EncryptedPayload)

PlaintextPayload = _reflection.GeneratedProtocolMessageType('PlaintextPayload', (_message.Message,), {
  'DESCRIPTOR' : _PLAINTEXTPAYLOAD,
  '__module__' : 'c3loc.proto.secbeacon_pb2'
  # @@protoc_insertion_point(class_scope:PlaintextPayload)
  })
_sym_db.RegisterMessage(PlaintextPayload)

SecBeaconReport = _reflection.GeneratedProtocolMessageType('SecBeaconReport', (_message.Message,), {
  'DESCRIPTOR' : _SECBEACONREPORT,
  '__module__' : 'c3loc.proto.secbeacon_pb2'
  # @@protoc_insertion_point(class_scope:SecBeaconReport)
  })
_sym_db.RegisterMessage(SecBeaconReport)


# @@protoc_insertion_point(module_scope)



import asyncio
from uuid import UUID

from typing import Any

from ..config import CONFIG
from .frame import Frame
from ..proto.common_pb2 import ListenerID
from ..proto.location_pb2 import IBeaconSummary, TempBeaconSummary, LABeaconSummary


class PacketType:
    """Enum emulation class."""

    ID = 0
    DATA = 1
    SECURE = 2
    RESPONSE = 3
    REQUEST = 4
    ID_REQUEST = 5
    HEALTH_CHECK = 6
    TEMPBEACON = 7
    LABEACON = 8


def batt_2032_pct(mvolts):
    batt_pct = 0

    if mvolts >= 3000:
        batt_pct = 100
    elif mvolts > 2900:
        batt_pct = 100 - ((3000 - mvolts) * 58) / 100
    elif mvolts > 2740:
        batt_pct = 42 - ((2900 - mvolts) * 24) / 160
    elif mvolts > 2440:
        batt_pct = 18 - ((2740 - mvolts) * 12) / 300
    elif mvolts > 2100:
        batt_pct = 6 - ((2440 - mvolts) * 6) / 340
    return batt_pct


class ListenerProtocol(object):
    callback = None
    db_pool = None

    def __init__(self, verbose=False):
        self._rx_buf = bytearray()
        self.transport = None
        self._frame = Frame()
        self.verbose = verbose
        if not self.db_pool:
            raise NotImplementedError("ListenerProtocol.db_pool must be set before instantiation")
        self.router = {
            PacketType.DATA: self.process_ibeacon,
            PacketType.LABEACON: self.process_relay_beacon,
        }
        self.ignored_types = {PacketType.HEALTH_CHECK, PacketType.SECURE, PacketType.TEMPBEACON}
        self._l_id = None
        self.listener_id = None
        self.peer_name = None

    def connection_made(self, transport) -> None:
        self.transport = transport
        self.peer_name = transport.get_extra_info('peername')[0]
        print("Connection from {}".format(self.peer_name))

    def eof_received(self):
        pass

    def data_received(self, data: bytes) -> None:
        self._frame.add_bytes(data)
        for msg in self._frame.get_messages():
            asyncio.create_task(self.process_packet(msg))

    def _print_reason(self, report):
        print("\t\tReason: {}".format(self._reason(report)))

    def _reason(self, report):
        reason = ""
        if report.reason in [IBeaconSummary.IBeaconReport.EventReason.ENTRY,
            TempBeaconSummary.TempBeaconReport.EventReason.ENTRY,
                             LABeaconSummary.LABeaconReport.EventReason.ENTRY]:
            reason = "ENTRY"
        if report.reason in [IBeaconSummary.IBeaconReport.EventReason.MOVE,
                             TempBeaconSummary.TempBeaconReport.EventReason.MOVE,
                             LABeaconSummary.LABeaconReport.EventReason.MOVE]:
            reason = "MOVE"
        if report.reason in [IBeaconSummary.IBeaconReport.EventReason.STATUS,
                             TempBeaconSummary.TempBeaconReport.EventReason.STATUS,
                             LABeaconSummary.LABeaconReport.EventReason.STATUS]:
            reason = "STATUS"
        if report.reason in [IBeaconSummary.IBeaconReport.EventReason.EXIT,
                             TempBeaconSummary.TempBeaconReport.EventReason.EXIT,
                             LABeaconSummary.LABeaconReport.EventReason.EXIT]:
            reason = "EXIT"
        return reason

    async def _upsert_listener(self, conn):
        await conn.execute('INSERT into listeners (id) VALUES ($1) '
                           'ON CONFLICT ON CONSTRAINT listeners_pkey '
                           'DO UPDATE SET last_seen = (now() at time zone \'utc\')',
                           self.listener_id)
        # Add a default zone for listeners
        zone_id = await self._select_zone_id(conn)
        if not zone_id:
            r = await conn.fetchrow('INSERT INTO zones (name) VALUES ($1) RETURNING id',
                                          f'Near Listener {self.listener_id}')
            zone_id = r[0]
            await conn.execute('UPDATE listeners SET zone_id = $1 WHERE id = $2',
                               zone_id, self.listener_id)

    async def _select_zone_id(self, conn):
        l = await conn.fetchrow('SELECT zone_id FROM listeners WHERE id = $1', self.listener_id)
        if not l:
            return None
        return l[0]

    async def process_packet(self, msg):
        if msg.type == PacketType.ID:
            self.process_id(msg)
            return

        if not self.listener_id:
            print(f"Warning: {self.peer_name} attempted non-ID packet before ID. Dropping connection")
            self.transport.close()
            return

        if msg.type in self.ignored_types:
            return

        conn = await self.db_pool.acquire()
        await self._upsert_listener(conn)  # Ensures listener in db and/or updates last_seen
        if msg.type not in self.router:
            print(f"Warning: No implementation for packet type: {msg.type}")
            return
        await self.router[msg.type](msg, conn)
        await self.db_pool.release(conn)

    def process_id(self, msg):
        l_id = ListenerID()
        l_id.ParseFromString(msg.data)
        self._l_id = l_id.listener_id
        self.listener_id = self._l_id.hex()
        print("{} identified as {} (version {})".format(
            self.peer_name, self._l_id.hex(), l_id.version))

    async def process_ibeacon(self, msg, conn):
        data = IBeaconSummary()
        data.ParseFromString(msg.data)

        # Convert to upsert when last_seen field is added?
        tag_upsert = await conn.prepare('INSERT into tags (uuid, major, minor, type)'
                                        'VALUES ($1, $2, $3, $4) ON CONFLICT ON CONSTRAINT uk_ibeacon '
                                        'DO UPDATE SET last_seen = (now() at time zone \'utc\')')
        tag_select = await conn.prepare('SELECT id FROM tags WHERE uuid = $1 AND major = $2 AND minor = $3')
        ibeacon_log = await conn.prepare('INSERT into log '
                                         '(tag_id, zone_id, listener_id, distance_cm, variance, reason)'
                                         'VALUES ($1, $2, $3, $4, $5, $6)')
        listener_zone = await self._select_zone_id(conn)
        for b in data.reports:
            uuid = UUID(bytes=b.uuid)
            tag_type = ''
            if uuid == CONFIG['LA_UUID']:
                tag_type = 'LocationAnchor'
            else:
                tag_type = 'iBeacon'
            await tag_upsert.fetch(uuid, b.major_num, b.minor_num, tag_type)
            t = await tag_select.fetchrow(uuid, b.major_num, b.minor_num)
            await ibeacon_log.fetch(t['id'], listener_zone, self.listener_id, b.distance_cm/100,
                                    round(b.variance/2000, 1), self._reason(b))

    async def process_relay_beacon(self, msg, conn):
        data = LABeaconSummary()
        data.ParseFromString(msg.data)
        # Convert to upsert when last_seen field is added?
        tag_upsert = await conn.prepare('INSERT into tags (mac, type)'
                                        'VALUES ($1, $2) ON CONFLICT ON CONSTRAINT uk_mac '
                                        'DO UPDATE SET last_seen = (now() at time zone \'utc\')')
        tag_select = await conn.prepare('SELECT id FROM tags WHERE mac = $1')
        anchor_upsert = await conn.prepare('INSERT into tags (uuid, major, minor, type) '
                                           'VALUES ($1, $2, $3, $4) ON CONFLICT ON CONSTRAINT uk_ibeacon '
                                           'DO UPDATE SET last_seen = (now() at time zone \'utc\') RETURNING id, zone_id')
        anchor_select = await conn.prepare('SELECT id, zone_id FROM tags WHERE uuid = $1 AND major = $2 AND minor = $3')
        labeacon_log = await conn.prepare('INSERT into log '
                                          '(tag_id, zone_id, listener_id, distance_cm,'
                                          'variance, anchor_id, anchor_dist, anchor_ts_delta, reason)'
                                          'VALUES ($1, $2, $3, $4, $5, $6, $7, $8, $9)')
        listener_zone = await self._select_zone_id(conn)
        for b in data.reports:
            b_id = bytes(reversed(b.beacon_id)).hex()
            await tag_upsert.fetch(b_id, 'SmartRelay')
            t = await tag_select.fetchrow(b_id)
            if b.nearest == 0:
                # No current anchor
                await labeacon_log.fetch(
                    t['id'], listener_zone, self.listener_id, b.distance_cm / 100,
                    round(b.variance / 2000, 1), None, None, None, self._reason(b))
                break
            # Reconstruct the iBeacon from the LA data
            major = b.nearest >> 16
            minor = b.nearest & 0xffff
            r = await anchor_upsert.fetchrow(CONFIG['LA_UUID'], major, minor, 'LocationAnchor')
            a_id, azone_id = r[0], r[1]
            if not azone_id:
                s = await conn.fetchrow('INSERT INTO zones (name) VALUES ($1) RETURNING id',
                                        f'Near LA {major}:{minor}')
                azone_id = s[0]
                await conn.execute('UPDATE tags SET zone_id = $1 WHERE id = $2',
                                   azone_id, a_id)
            anchor = await anchor_select.fetchrow(CONFIG['LA_UUID'], major, minor)
            await labeacon_log.fetch(
                t['id'], anchor['zone_id'], self.listener_id, b.distance_cm / 100, round(b.variance / 2000, 1),
                anchor['id'], b.nearest_dist/10, b.nearest_delta, self._reason(b))

    async def process_health_check(self, msg, conn):
        pass

    def error_received(self, exc: Exception) -> None:
        print(f'Error received: {exc}')

    def connection_lost(self, exc: Exception) -> None:
        print("{} has disconnected".format(self.peer_name))



from aiohttp import web
from aiohttp_cors import CorsViewMixin
import asyncpg

from .schemas import prox_schema
from .views import paginate_query
from ..config import CONFIG

THE_QUERY = \
"""
SELECT DISTINCT ON (upper.tag_id) coalesce(anchor_dist, distance_cm) as distance,
                                zones.name as zone_name,
                                tags.name as tag_name,
                                tags.type as tag_type,
                                tags.last_seen as last_seen,
                                upper.anchor_id, upper.tag_id, upper.zone_id from log upper
LEFT JOIN zones on zones.id = upper.zone_id
LEFT JOIN (SELECT * FROM tags t) tags on tags.id = upper.tag_id
WHERE ts > tags.last_seen - interval '1min'
ORDER BY upper.tag_id, distance, upper.ts DESC
"""

THE_TYPE_QUERY = \
"""
SELECT DISTINCT ON (upper.tag_id) coalesce(anchor_dist, distance_cm) as distance,
                                zones.name as zone_name,
                                tags.name as tag_name,
                                tags.type as tag_type,
                                tags.last_seen as last_seen,
                                upper.anchor_id, upper.tag_id, upper.zone_id from log upper
LEFT JOIN zones on zones.id = upper.zone_id
LEFT JOIN (SELECT * FROM tags t) tags on tags.id = upper.tag_id
WHERE ts > tags.last_seen - interval '1min' AND tags.type = $1
ORDER BY upper.tag_id, distance, upper.ts DESC
"""

THE_GROUP_QUERY = \
"""
SELECT DISTINCT ON (upper.tag_id) coalesce(anchor_dist, distance_cm) as distance,
                                zones.name as zone_name,
                                tags.name as tag_name,
                                tags.type as tag_type,
                                tags.last_seen as last_seen,
                                upper.anchor_id, upper.tag_id, upper.zone_id from log upper
LEFT JOIN zones on zones.id = upper.zone_id
LEFT JOIN (SELECT * FROM tags t) tags on tags.id = upper.tag_id
WHERE ts > tags.last_seen - interval '1min' AND tags.group_id = $1
ORDER BY upper.tag_id, distance, upper.ts DESC
"""
# """SELECT zones.name as zone_name, tags.name as tag_name, a.*
# FROM (SELECT log.tag_id, zone_id, anchor_dist as distance, log.ts, anchor_id from log
#     JOIN (SELECT tag_id, max(ts) max_ts from log WHERE anchor_id IS NOT NULL GROUP BY tag_id) temp
#         ON log.tag_id = temp.tag_id AND log.ts = temp.max_ts
# UNION SELECT DISTINCT ON (log.tag_id) log.tag_id, zone_id, distance_cm as distance, log.ts, anchor_id from log
#     JOIN (SELECT tag_id, min(distance_cm), ts from log
#           WHERE anchor_id IS NULL and ts > ((now() at time zone 'utc') - interval '1 min') GROUP BY tag_id, ts ORDER BY ts DESC) temp
#     ON log.tag_id = temp.tag_id AND log.ts = temp.ts) a
# LEFT JOIN zones on zone_id = zones.id
# LEFT JOIN tags on tag_id = tags.id"""


def process_prox_results(results):
    output = []
    for r in results:
        anchor_id = r['anchor_id']
        zone_id = r['zone_id']
        d_r = dict(r)
        d_r['links'] = {
            "anchor": f"/api/tags/{anchor_id}" if anchor_id else None,
            "zone": f"/api/zones/{zone_id}" if zone_id else None,
            "tag": f"/api/tags/{r['tag_id']}"
        }
        output.append(d_r)
    return output


class ProximityView(web.View, CorsViewMixin):
    async def get(self):
        if 'type' in self.request.query:
            t = self.request.query['type']
            query = (THE_TYPE_QUERY, t)
        elif 'group_id' in self.request.query:
            g = self.request.query['group_id']
            try:
                query = (THE_GROUP_QUERY, int(g))
            except ValueError:
                raise web.HTTPBadRequest(text='Invalid group id')
        else:
            query = (THE_QUERY,)
        query = paginate_query(self.request, query)
        async with self.request.app['db_pool'].acquire() as conn:
            try:
                results = await conn.fetch(*query)
                print(process_prox_results(results))
            except asyncpg.exceptions.InvalidTextRepresentationError:
                raise web.HTTPBadRequest(text='Bad tag type in query')
            return web.json_response(prox_schema.dump(process_prox_results(results)))



import os
from uuid import UUID

CONFIG = {
    'DB_NAME': os.getenv('DB_NAME', ''),
    'DB_USER': os.getenv('DB_USER', 'c3loc'),
    'DB_PASSWORD': os.getenv('DB_PASSWORD', 'c3letmein'),
    'DB_HOST': os.getenv('DB_HOST', ''),
    'LA_UUID': os.getenv('LA_UUID', UUID(bytes=b'c3wirelesslocanc')),
    'API_RESULT_LIMIT': int(os.getenv('API_RESULT_LIMIT', 100)),
}



import json

from aiohttp import web

from .schemas import groups_schema, group_schema
from .views import ValidatingView, paginate_query


class GroupsView(ValidatingView):
    async def get(self):
        async with self.request.app['db_pool'].acquire() as conn:
            query = ('SELECT * from groups', )
            tags = await conn.fetch(*paginate_query(self.request, query))
            return web.json_response(groups_schema.dump(tags))

    async def post(self):
        body = await self._valid_json(self.request)
        new_z = self._validate(group_schema, body)

        if 'attrs' not in new_z:
            new_z['attrs'] = {}

        async with self.request.app['db_pool'].acquire() as conn:
            await conn.execute('INSERT INTO groups (name, attrs) VALUES ($1, $2)',
                         new_z['name'], json.dumps(new_z['attrs']))
        raise web.HTTPCreated


class GroupView(ValidatingView):
    async def get(self):
        z_id = int(self.request.match_info['id'])
        async with self.request.app['db_pool'].acquire() as conn:
            l = await conn.fetchrow('SELECT * FROM groups WHERE id = $1', z_id)
            if not l:
                raise web.HTTPNotFound
            return web.json_response(group_schema.dump(l))

    async def patch(self):
        z_id = int(self.request.match_info['id'])
        async with self.request.app['db_pool'].acquire() as conn:
            l_row = await conn.fetchrow('SELECT name, attrs FROM groups WHERE id = $1', z_id)
            if not l_row:
                raise web.HTTPNotFound
            group = group_schema.load(dict(l_row.items()))
            body = await self._valid_json(self.request)
            new = self._validate(group_schema, body)
            group.update(new)
            await conn.execute('UPDATE groups SET name = $1, attrs = $2 WHERE id = $3',
                               group['name'], json.dumps(group['attrs']), z_id)
        raise web.HTTPNoContent

    async def delete(self):
        z_id = int(self.request.match_info['id'])
        async with self.request.app['db_pool'].acquire() as conn:
            l_row = await conn.fetchrow('SELECT name, attrs FROM groups WHERE id = $1', z_id)
            if not l_row:
                raise web.HTTPNotFound
            await conn.execute('DELETE from groups where id = $1', z_id)
        raise web.HTTPNoContent



import json
from collections import OrderedDict

from aiohttp import web
import asyncpg
from marshmallow import EXCLUDE

from .schemas import ibeacon_post_schema, macbeacon_post_schema, tag_patch_schema
from .views import ValidatingView, paginate_query


def tag_schema_dump(row):
    if row['type'] == 'iBeacon':
        return ibeacon_post_schema.dump(row)
    return macbeacon_post_schema.dump(row)


class TagsView(ValidatingView):
    async def get(self):
        async with self.request.app['db_pool'].acquire() as conn:
            query = None
            if 'type' in self.request.query:
                query = ('SELECT * from tags WHERE type = $1',
                         self.request.query['type'])
            elif 'group_id' in self.request.query:
                try:
                    query = ('SELECT * from tags WHERE group_id = $1',
                             int(self.request.query['group_id']))
                except ValueError:
                    raise web.HTTPBadRequest(text='bad group_id')
            else:
                query = ('SELECT * from tags', )
            query = paginate_query(self.request, query)
            try:
                tags = await conn.fetch(*query)
            except asyncpg.exceptions.InvalidTextRepresentationError:
                raise web.HTTPBadRequest(text='Invalid tag type')
            return web.json_response([tag_schema_dump(r) for r in tags if r is not None])

    async def post(self):
        body = await self._valid_json(self.request)

        if 'type' not in body:
            raise web.HTTPBadRequest(reason='Type field is required for post')

        new_t = {}
        if body['type'] in {'iBeacon', 'LocationAnchor'}:
            new_t.update(self._validate(ibeacon_post_schema, body))
        elif body['type'] == 'SmartRelay':
            new_t.update(self._validate(macbeacon_post_schema, body))
        else:
            raise web.HTTPBadRequest(reason=f'Unknown tag type {body["type"]}')

        default = OrderedDict([
            ('mac', None),
            ('uuid', None),
            ('major', None),
            ('minor', None),
            ('zone_id', None),
            ('name', None),
            ('attrs', "{}"),
        ])
        default.update(new_t)
        del default['type']

        async with self.request.app['db_pool'].acquire() as conn:
            try:
                await conn.execute('INSERT INTO tags (mac, uuid, major, minor, zone_id, name, attrs) VALUES '
                                   '($1, $2, $3, $4, $5, $6, $7)', *default.values())
            except asyncpg.exceptions.UniqueViolationError as e:
                raise web.HTTPConflict(reason=e.detail)
        raise web.HTTPCreated


class TagView(ValidatingView):
    async def get(self):
        t_id = int(self.request.match_info['id'])
        async with self.request.app['db_pool'].acquire() as conn:
            l = await conn.fetchrow('SELECT * FROM tags t WHERE id = $1', t_id)
            if not l:
                raise web.HTTPNotFound
            return web.json_response(tag_schema_dump(l))

    async def patch(self):
        t_id = int(self.request.match_info['id'])
        async with self.request.app['db_pool'].acquire() as conn:
            row = await conn.fetchrow('SELECT * FROM tags t WHERE id = $1', t_id)
            if not row:
                raise web.HTTPNotFound
            tag = dict(row.items())
            body = await self._valid_json(self.request)
            updates = self._validate(tag_patch_schema, body)
            if 'type' in updates:
                del body['type']  # Ensure that we are validating against existing type
            tag.update(updates)
            if tag['type'] in {'iBeacon', 'LocationAnchor'}:
                self._validate(ibeacon_post_schema, tag, unknown=EXCLUDE)
            else:
                self._validate(macbeacon_post_schema, tag, unknown=EXCLUDE)
            await conn.execute('UPDATE tags SET name = $1, mac = $2, uuid = $3, major = $4, '
                               'minor = $5, attrs = $6, zone_id = $7  WHERE id = $8',
                               tag['name'], tag['mac'], tag['uuid'], tag['major'], tag['minor'],
                               json.dumps(tag['attrs']), tag['zone_id'], t_id)
        raise web.HTTPNoContent

    async def delete(self):
        t_id = int(self.request.match_info['id'])
        async with self.request.app['db_pool'].acquire() as conn:
            await conn.execute('DELETE from tags where id = $1', t_id)
        raise web.HTTPNoContent



from setuptools import find_packages, setup
setup(
    name='c3loc',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version='0.3.4',
    description='C3 Enhanced Proximity Location Services',
    url='https://gitlab.com/C3Wireless/c3loc',
    author='C3 Wireless',
    author_email='support@c3wireless.com',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    keywords='c3 c3wireless btle beacon ibeacon',
    packages=find_packages(where='src'),
    package_dir={"": "src"},
    install_requires=[
        'asyncpg',
        'aiohttp',
        'aiohttp-cors',
        'aiohttp-jinja2',
        'automat',
        'click',
        'marshmallow',
        'protobuf',
        'python-lzo',
        'pycryptodome',
    ],
    extras_require={
        'test': ['tox']
    },
    entry_points={
        'console_scripts': [
            'c3loc_ingest=c3loc.ingest.main:main',
            'c3loc_api=c3loc.api.main:main',
        ],
    },
    package_data={'c3loc': ['static/*', 'templates/*']}
)



import asyncpg

from .config import CONFIG


async def init_db_pool():
    db_url = f"postgres://{CONFIG['DB_USER']}:{CONFIG['DB_PASSWORD']}@{CONFIG['DB_HOST']}/{CONFIG['DB_NAME']}"
    db_url_redacted = f"postgres://{CONFIG['DB_USER']}:*****@{CONFIG['DB_HOST']}"
    print(f"Connecting to {db_url_redacted}")
    pool = await asyncpg.create_pool(dsn=db_url)
    return pool


async def init_db(pool, reset_db):
    tables = ['zones', 'tags', 'listener', 'log']
    async with pool.acquire() as connection:
        if reset_db:
            for table in tables:
                await connection.execute(f'DROP TABLE IF EXISTS {table} CASCADE')
        async with connection.transaction():
            print('\tEnsuring schema')
            await connection.execute(DB_CREATE_STR)


DB_CREATE_STR = """
CREATE TABLE IF NOT EXISTS zones (
    id serial,
    name varchar,
    attrs jsonb,
    PRIMARY KEY(id));

CREATE TABLE IF NOT EXISTS groups (
    id serial,
    name varchar,
    attrs jsonb,
    PRIMARY KEY(id));

CREATE TABLE IF NOT EXISTS listeners (
    id varchar not null,
    name varchar,
    zone_id int,
    last_seen timestamp without time zone DEFAULT (now() at time zone 'utc'),
    attrs jsonb,
    PRIMARY KEY(id),
    UNIQUE(name),
    CONSTRAINT fk_zone
        FOREIGN KEY(zone_id)
        REFERENCES zones(id)
        ON DELETE SET NULL);

DO $$ BEGIN
    CREATE TYPE tag_type as ENUM ('iBeacon', 'SmartRelay', 'LocationAnchor');
EXCEPTION
    WHEN duplicate_object THEN null;
END $$;

CREATE TABLE IF NOT EXISTS tags (
    id serial,
    name varchar,
    mac macaddr,
    uuid uuid,
    major int4,
    minor int4,
    type tag_type NOT NULL,
    attrs jsonb,
    zone_id integer,
    last_seen timestamp without time zone DEFAULT (now() at time zone 'utc'),
    group_id integer,
    PRIMARY KEY (id),
    CONSTRAINT uk_ibeacon
    UNIQUE (uuid, major, minor),
    CONSTRAINT uk_mac
    UNIQUE (mac),
    CONSTRAINT fk_zone
                  FOREIGN KEY (zone_id)
                  REFERENCES zones(id)
                  ON DELETE SET NULL,
    CONSTRAINT fk_group
                  FOREIGN KEY (group_id)
                  REFERENCES groups(id)
                  ON DELETE SET NULL
);

CREATE OR REPLACE FUNCTION def_tag_name() RETURNS TRIGGER LANGUAGE plpgsql AS $$
BEGIN
    IF NEW.name IS NOT NULL OR NEW.name != ''
       THEN RETURN NEW;
    ELSIF NEW.type = 'iBeacon' THEN
        NEW.name := 'iBeacon ' || NEW.uuid || ':' || NEW.major || ':' || NEW.minor;
    ELSIF NEW.type = 'LocationAnchor' THEN
            NEW.name := 'Location Anchor ' || NEW.major || ':' || NEW.minor;
    ELSIF NEW.type = 'SmartRelay' THEN
            NEW.name := 'SmartRelay ' || NEW.mac;
    END IF;
    return NEW;
END
$$;

DO $$ BEGIN
    CREATE TRIGGER tags_set_def_name
    BEFORE INSERT OR UPDATE ON tags
    FOR EACH ROW
    EXECUTE PROCEDURE def_tag_name();
EXCEPTION
    WHEN duplicate_object THEN null;
END $$;

DO $$ BEGIN
    CREATE TYPE report_reason as ENUM ('ENTRY', 'MOVE', 'STATUS', 'EXIT');
EXCEPTION
    WHEN duplicate_object THEN null;
END $$;

CREATE TABLE IF NOT EXISTS log (
    tag_id int,
    zone_id int,
    ts timestamp without time zone default (now() at time zone 'utc'),
    distance_cm numeric(4,2),
    variance real,
    listener_id varchar,
    data jsonb,
    anchor_dist numeric(4,2),
    anchor_ts_delta int,
    anchor_id int,
    reason report_reason,
    CONSTRAINT fk_tag
                 FOREIGN KEY (tag_id)
                 REFERENCES tags(id)
                 ON DELETE CASCADE,
    CONSTRAINT fk_zone
                 FOREIGN KEY (zone_id)
                 REFERENCES zones(id)
                 ON DELETE SET NULL,
    CONSTRAINT fk_anchor
                 FOREIGN KEY (anchor_id)
                 REFERENCES tags(id)
                 ON DELETE SET NULL,
    CONSTRAINT fk_listener
                 FOREIGN KEY (listener_id)
                 REFERENCES listeners(id)
                 ON DELETE SET NULL
);

CREATE INDEX IF NOT EXISTS idx_log_tag_id ON log (tag_id);
CREATE INDEX IF NOT EXISTS idx_log_anchor_id ON log (anchor_id);
CREATE INDEX IF NOT EXISTS idx_log_zone_id ON log (zone_id);
CREATE INDEX IF NOT EXISTS idx_log_ts ON log (ts DESC);
"""
